@@ -4,6 +4,7 @@
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE TypeOperators #-}
 -- | This module provides the core widget combinators and rendering
 -- routines. Everything this library does is in terms of these basic
 -- primitives.
@@ -22,6 +23,7 @@ module Brick.Widgets.Core
   , hyperlink
 
   -- * Padding
+  , Padding(..)
   , padLeft
   , padRight
   , padTop
@@ -48,6 +50,7 @@ module Brick.Widgets.Core
   , modifyDefAttr
   , withAttr
   , forceAttr
+  , forceAttrAllowStyle
   , overrideAttr
   , updateAttrMap
 
@@ -64,8 +67,9 @@ module Brick.Widgets.Core
   -- * Naming
   , Named(..)
 
-  -- * Translation
+  -- * Translation and positioning
   , translateBy
+  , relativeTo
 
   -- * Cropping
   , cropLeftBy
@@ -120,13 +124,12 @@ import Data.Monoid ((<>))
 
 import Lens.Micro ((^.), (.~), (&), (%~), to, _1, _2, each, to, Lens')
 import Lens.Micro.Mtl (use, (%=))
-import Control.Monad ((>=>),when)
-import Control.Monad.Trans.State.Lazy
-import Control.Monad.Trans.Reader
-import Control.Monad.Trans.Class (lift)
+import Control.Monad
+import Control.Monad.State.Strict
+import Control.Monad.Reader
 import qualified Data.Foldable as F
+import Data.Traversable (for)
 import qualified Data.Text as T
-import qualified Data.DList as DL
 import qualified Data.Map as M
 import qualified Data.Set as S
 import qualified Data.IMap as I
@@ -193,13 +196,13 @@ separateBorders p = Widget (hSize p) (vSize p) $
 --
 -- Frozen borders cannot be thawed.
 freezeBorders :: Widget n -> Widget n
-freezeBorders p = Widget (hSize p) (vSize p) $ (bordersL .~ BM.empty) <$> render p
+freezeBorders p = Widget (hSize p) (vSize p) $ (bordersL %~ BM.clear) <$> render p
 
 -- | The empty widget.
 emptyWidget :: Widget n
 emptyWidget = raw V.emptyImage
 
--- | Add an offset to all cursor locations, visbility requests, and
+-- | Add an offset to all cursor locations, visibility requests, and
 -- extents in the specified rendering result. This function is critical
 -- for maintaining correctness in the rendering results as they are
 -- processed successively by box layouts and other wrapping combinators,
@@ -268,17 +271,6 @@ addCursorOffset off r =
 unrestricted :: Int
 unrestricted = 100000
 
--- | Take a substring capable of fitting into the number of specified
--- columns. This function takes character column widths into
--- consideration.
-takeColumns :: Int -> String -> String
-takeColumns _ "" = ""
-takeColumns numCols (c:cs) =
-    let w = V.safeWcwidth c
-    in if w > numCols
-       then []
-       else c : takeColumns (numCols - w) cs
-
 -- | Make a widget from a string, but wrap the words in the input's
 -- lines at the available width using the default wrapping settings. The
 -- input string should not contain escape sequences or carriage returns.
@@ -296,9 +288,6 @@ strWrap = strWrapWith defaultWrapSettings
 strWrapWith :: WrapSettings -> String -> Widget n
 strWrapWith settings t = txtWrapWith settings $ T.pack t
 
-safeTextWidth :: T.Text -> Int
-safeTextWidth = V.safeWcswidth . T.unpack
-
 -- | Make a widget from text, but wrap the words in the input's lines at
 -- the available width using the default wrapping settings. The input
 -- text should not contain escape sequences or carriage returns.
@@ -322,15 +311,15 @@ txtWrapWith settings s =
       case force theLines of
           [] -> return emptyResult
           multiple ->
-              let maxLength = maximum $ safeTextWidth <$> multiple
+              let maxLength = maximum $ textWidth <$> multiple
                   padding = V.charFill (c^.attrL) ' ' (c^.availWidthL - maxLength) (length lineImgs)
                   lineImgs = lineImg <$> multiple
                   lineImg lStr = V.text' (c^.attrL)
-                                   (lStr <> T.replicate (maxLength - safeTextWidth lStr) " ")
+                                   (lStr <> T.replicate (maxLength - textWidth lStr) " ")
               in return $ emptyResult & imageL .~ (V.horizCat [V.vertCat lineImgs, padding])
 
--- | Build a widget from a 'String'. Breaks newlines up and space-pads
--- short lines out to the length of the longest line.
+-- | Build a widget from a 'String'. Behaves the same as 'txt' when the
+-- input contains multiple lines.
 --
 -- The input string must not contain tab characters. If it does,
 -- interface corruption will result since the terminal will likely
@@ -338,25 +327,10 @@ txtWrapWith settings s =
 -- replace tabs with the appropriate number of spaces as desired. The
 -- input string should not contain escape sequences or carriage returns.
 str :: String -> Widget n
-str s =
-    Widget Fixed Fixed $ do
-      c <- getContext
-      let theLines = fixEmpty <$> (dropUnused . lines) s
-          fixEmpty :: String -> String
-          fixEmpty [] = " "
-          fixEmpty l = l
-          dropUnused l = takeColumns (availWidth c) <$> take (availHeight c) l
-      case force theLines of
-          [] -> return emptyResult
-          [one] -> return $ emptyResult & imageL .~ (V.string (c^.attrL) one)
-          multiple ->
-              let maxLength = maximum $ V.safeWcswidth <$> multiple
-                  lineImgs = lineImg <$> multiple
-                  lineImg lStr = V.string (c^.attrL) (lStr ++ replicate (maxLength - V.safeWcswidth lStr) ' ')
-              in return $ emptyResult & imageL .~ (V.vertCat lineImgs)
+str = txt . T.pack
 
--- | Build a widget from a 'T.Text' value. Behaves the same as 'str'
--- when the input contains multiple lines.
+-- | Build a widget from a 'T.Text' value. Breaks newlines up and
+-- space-pads short lines out to the length of the longest line.
 --
 -- The input string must not contain tab characters. If it does,
 -- interface corruption will result since the terminal will likely
@@ -364,7 +338,44 @@ str s =
 -- replace tabs with the appropriate number of spaces as desired. The
 -- input text should not contain escape sequences or carriage returns.
 txt :: T.Text -> Widget n
-txt = str . T.unpack
+txt s =
+    -- Althoguh vty Image uses lazy Text internally, using lazy text at this
+    -- level may not be an improvement.  Indeed it can be much worse, due
+    -- the overhead of lazy Text being significant compared to the typically
+    -- short string content used to compose UIs.
+    Widget Fixed Fixed $ do
+        c <- getContext
+        let theLines = fixEmpty <$> (dropUnused . T.lines) s
+            fixEmpty l = if T.null l then T.singleton ' ' else l
+            dropUnused l = takeColumnsT (availWidth c) <$> take (availHeight c) l
+        pure $ case theLines of
+            [] -> emptyResult
+            [one] -> emptyResult & imageL .~ (V.text' (c^.attrL) one)
+            multiple ->
+                let maxLength = maximum $ V.safeWctwidth <$> multiple
+                    lineImgs = lineImg <$> multiple
+                    lineImg lStr = V.text' (c^.attrL)
+                        (lStr <> T.replicate (maxLength - V.safeWctwidth lStr) (T.singleton ' '))
+                in emptyResult & imageL .~ (V.vertCat lineImgs)
+
+-- | Take up to the given width, having regard to character width.
+takeColumnsT :: Int -> T.Text -> T.Text
+takeColumnsT w s = T.take (fst $ T.foldl' f (0,0) s) s
+    where
+    -- The accumulator value is (index in Text value, width of Text so far)
+    f (i,z) c
+        -- Width was previously exceeded; continue with same values.
+        | z < 0                   = (i, z)
+        -- Width exceeded.  Signal this with z = -1.  Index will no longer be
+        -- incremented.
+        --
+        -- Why not short circuit (e.g. using foldlM construction)?
+        -- Because in the typical case, the Either allocation costs exceed
+        -- any benefits.  The pathological case, string length >> width, is
+        -- probably rare.
+        | z + V.safeWcwidth c > w = (i, -1)
+        -- Width not yet exceeded.  Increment index and add character width.
+        | otherwise               = (i + 1, z + V.safeWcwidth c)
 
 -- | Hyperlink the given widget to the specified URL. Not all terminal
 -- emulators support this. In those that don't, this should have no
@@ -376,6 +387,12 @@ hyperlink url p =
         let attr = (c^.attrL) `V.withURL` url
         withReaderT (ctxAttrMapL %~ setDefaultAttr attr) (render p)
 
+-- | The type of padding.
+data Padding = Pad Int
+             -- ^ Pad by the specified number of rows or columns.
+             | Max
+             -- ^ Pad up to the number of available rows or columns.
+
 -- | Pad the specified widget on the left. If max padding is used, this
 -- grows greedily horizontally; otherwise it defers to the padded
 -- widget.
@@ -470,6 +487,19 @@ fill ch =
 -- in the specified order (uppermost first). Defers growth policies to
 -- the growth policies of the contained widgets (if any are greedy, so
 -- is the box).
+--
+-- Allocates space to 'Fixed' elements first and 'Greedy' elements
+-- second. For example, if a 'vBox' contains three elements @A@, @B@,
+-- and @C@, and if @A@ and @B@ are 'Fixed', then 'vBox' first renders
+-- @A@ and @B@. Suppose those two take up 10 rows total, and the 'vBox'
+-- was given 50 rows. This means 'vBox' then allocates the remaining
+-- 40 rows to @C@. If, on the other hand, @A@ and @B@ take up 50 rows
+-- together, @C@ will not be rendered at all.
+--
+-- If all elements are 'Greedy', 'vBox' allocates the available height
+-- evenly among the elements. So, for example, if a 'vBox' is rendered
+-- in 90 rows and has three 'Greedy' elements, each element will be
+-- allocated 30 rows.
 {-# NOINLINE vBox #-}
 vBox :: [Widget n] -> Widget n
 vBox [] = emptyWidget
@@ -480,6 +510,19 @@ vBox pairs = renderBox vBoxRenderer pairs
 -- in the specified order (leftmost first). Defers growth policies to
 -- the growth policies of the contained widgets (if any are greedy, so
 -- is the box).
+--
+-- Allocates space to 'Fixed' elements first and 'Greedy' elements
+-- second. For example, if an 'hBox' contains three elements @A@, @B@,
+-- and @C@, and if @A@ and @B@ are 'Fixed', then 'hBox' first renders
+-- @A@ and @B@. Suppose those two take up 10 columns total, and the
+-- 'hBox' was given 50 columns. This means 'hBox' then allocates the
+-- remaining 40 columns to @C@. If, on the other hand, @A@ and @B@ take
+-- up 50 columns together, @C@ will not be rendered at all.
+--
+-- If all elements are 'Greedy', 'hBox' allocates the available width
+-- evenly among the elements. So, for example, if an 'hBox' is rendered
+-- in 90 columns and has three 'Greedy' elements, each element will be
+-- allocated 30 columns.
 {-# NOINLINE hBox #-}
 hBox :: [Widget n] -> Widget n
 hBox [] = emptyWidget
@@ -502,7 +545,6 @@ data BoxRenderer n =
                 , imagePrimary :: V.Image -> Int
                 , imageSecondary :: V.Image -> Int
                 , limitPrimary :: Int -> Widget n -> Widget n
-                , limitSecondary :: Int -> Widget n -> Widget n
                 , primaryWidgetSize :: Widget n -> Size
                 , concatenatePrimary :: [V.Image] -> V.Image
                 , concatenateSecondary :: [V.Image] -> V.Image
@@ -528,7 +570,6 @@ vBoxRenderer =
                 , imagePrimary = V.imageHeight
                 , imageSecondary = V.imageWidth
                 , limitPrimary = vLimit
-                , limitSecondary = hLimit
                 , primaryWidgetSize = vSize
                 , concatenatePrimary = V.vertCat
                 , concatenateSecondary = V.horizCat
@@ -556,7 +597,6 @@ hBoxRenderer =
                 , imagePrimary = V.imageWidth
                 , imageSecondary = V.imageHeight
                 , limitPrimary = hLimit
-                , limitSecondary = vLimit
                 , primaryWidgetSize = hSize
                 , concatenatePrimary = V.horizCat
                 , concatenateSecondary = V.vertCat
@@ -633,45 +673,33 @@ renderBox br ws =
           (his, lows) = partition (\p -> (primaryWidgetSize br $ snd p) == Fixed)
                         pairsIndexed
 
-      let availPrimary = c^.(contextPrimary br)
-          availSecondary = c^.(contextSecondary br)
+          renderHi prim = do
+            remainingPrimary <- get
+            result <- lift $ render $ limitPrimary br remainingPrimary prim
+            result <$ (put $! remainingPrimary - (result^.imageL.(to $ imagePrimary br)))
 
-          renderHis _ prev [] = return $ DL.toList prev
-          renderHis remainingPrimary prev ((i, prim):rest) = do
-              result <- render $ limitPrimary br remainingPrimary
-                               $ limitSecondary br availSecondary
-                               $ cropToContext prim
-              renderHis (remainingPrimary - (result^.imageL.(to $ imagePrimary br)))
-                        (DL.snoc prev (i, result)) rest
-
-      renderedHis <- renderHis availPrimary DL.empty his
+      (renderedHis, remainingPrimary) <-
+        runStateT (traverse (traverse renderHi) his) (c ^. contextPrimary br)
 
       renderedLows <- case lows of
           [] -> return []
           ls -> do
-              let remainingPrimary = c^.(contextPrimary br) -
-                                     (sum $ (^._2.imageL.(to $ imagePrimary br)) <$> renderedHis)
-                  primaryPerLow = remainingPrimary `div` length ls
+              let primaryPerLow = remainingPrimary `div` length ls
                   rest = remainingPrimary - (primaryPerLow * length ls)
-                  secondaryPerLow = c^.(contextSecondary br)
                   primaries = replicate rest (primaryPerLow + 1) <>
                               replicate (length ls - rest) primaryPerLow
 
-              let renderLow ((i, prim), pri) =
-                      (i,) <$> (render $ limitPrimary br pri
-                                       $ limitSecondary br secondaryPerLow
-                                       $ cropToContext prim)
+              let renderLow ((i, prim), pri) = (i,) <$> render (limitPrimary br pri prim)
 
               if remainingPrimary > 0 then mapM renderLow (zip ls primaries) else return []
 
       let rendered = sortBy (compare `DF.on` fst) $ renderedHis ++ renderedLows
           allResults = snd <$> rendered
           allImages = (^.imageL) <$> allResults
-          allPrimaries = imagePrimary br <$> allImages
-          allTranslatedResults = (flip map) (zip [0..] allResults) $ \(i, result) ->
-              let off = locationFromOffset br offPrimary
-                  offPrimary = sum $ take i allPrimaries
-              in addResultOffset off result
+          allTranslatedResults = flip evalState 0 $ for allResults $ \result -> do
+              offPrimary <- get
+              put $ offPrimary + (result ^. imageL . to (imagePrimary br))
+              pure $ addResultOffset (locationFromOffset br offPrimary) result
           -- Determine the secondary dimension value to pad to. In a
           -- vertical box we want all images to be the same width to
           -- avoid attribute over-runs or blank spaces with the wrong
@@ -685,9 +713,9 @@ renderBox br ws =
           paddedImages = padImage <$> rewrittenImages
 
       cropResultToContext $ Result (concatenatePrimary br paddedImages)
-                            (concat $ cursors <$> allTranslatedResults)
-                            (concat $ visibilityRequests <$> allTranslatedResults)
-                            (concat $ extents <$> allTranslatedResults)
+                            (concatMap cursors allTranslatedResults)
+                            (concatMap visibilityRequests allTranslatedResults)
+                            (concatMap extents allTranslatedResults)
                             newBorders
 
 catDynBorder
@@ -809,9 +837,11 @@ rewriteImage br (loRewrite, hiRewrite) old = rewriteHi . rewriteLo $ old where
 -- growth of otherwise-greedy widgets. This is non-greedy horizontally
 -- and defers to the limited widget vertically.
 hLimit :: Int -> Widget n -> Widget n
-hLimit w p =
-    Widget Fixed (vSize p) $
-      withReaderT (availWidthL %~ (min w)) $ render $ cropToContext p
+hLimit w p
+    | w <= 0 = emptyWidget
+    | otherwise =
+        Widget Fixed (vSize p) $
+          withReaderT (availWidthL %~ (min w)) $ render $ cropToContext p
 
 -- | Limit the space available to the specified widget to the specified
 -- percentage of available width, as a value between 0 and 100
@@ -820,22 +850,26 @@ hLimit w p =
 -- growth of otherwise-greedy widgets. This is non-greedy horizontally
 -- and defers to the limited widget vertically.
 hLimitPercent :: Int -> Widget n -> Widget n
-hLimitPercent w' p =
-    Widget Fixed (vSize p) $ do
-      let w = clamp 0 100 w'
-      ctx <- getContext
-      let usableWidth = ctx^.availWidthL
-          widgetWidth = round (toRational usableWidth * (toRational w / 100))
-      withReaderT (availWidthL %~ (min widgetWidth)) $ render $ cropToContext p
+hLimitPercent w' p
+    | w' <= 0 = emptyWidget
+    | otherwise =
+        Widget Fixed (vSize p) $ do
+          let w = clamp 0 100 w'
+          ctx <- getContext
+          let usableWidth = ctx^.availWidthL
+              widgetWidth = round (toRational usableWidth * (toRational w / 100))
+          withReaderT (availWidthL %~ (min widgetWidth)) $ render $ cropToContext p
 
 -- | Limit the space available to the specified widget to the specified
 -- number of rows. This is important for constraining the vertical
 -- growth of otherwise-greedy widgets. This is non-greedy vertically and
 -- defers to the limited widget horizontally.
 vLimit :: Int -> Widget n -> Widget n
-vLimit h p =
-    Widget (hSize p) Fixed $
-      withReaderT (availHeightL %~ (min h)) $ render $ cropToContext p
+vLimit h p
+    | h <= 0 = emptyWidget
+    | otherwise =
+        Widget (hSize p) Fixed $
+          withReaderT (availHeightL %~ (min h)) $ render $ cropToContext p
 
 -- | Limit the space available to the specified widget to the specified
 -- percentage of available height, as a value between 0 and 100
@@ -844,31 +878,36 @@ vLimit h p =
 -- growth of otherwise-greedy widgets. This is non-greedy vertically and
 -- defers to the limited widget horizontally.
 vLimitPercent :: Int -> Widget n -> Widget n
-vLimitPercent h' p =
-    Widget (hSize p) Fixed $ do
-      let h = clamp 0 100 h'
-      ctx <- getContext
-      let usableHeight = ctx^.availHeightL
-          widgetHeight = round (toRational usableHeight * (toRational h / 100))
-      withReaderT (availHeightL %~ (min widgetHeight)) $ render $ cropToContext p
+vLimitPercent h' p
+    | h' <= 0 = emptyWidget
+    | otherwise =
+        Widget (hSize p) Fixed $ do
+          let h = clamp 0 100 h'
+          ctx <- getContext
+          let usableHeight = ctx^.availHeightL
+              widgetHeight = round (toRational usableHeight * (toRational h / 100))
+          withReaderT (availHeightL %~ (min widgetHeight)) $ render $ cropToContext p
 
 -- | Set the rendering context height and width for this widget. This
 -- is useful for relaxing the rendering size constraints on e.g. layer
 -- widgets where cropping to the screen size is undesirable.
 setAvailableSize :: (Int, Int) -> Widget n -> Widget n
-setAvailableSize (w, h) p =
-    Widget Fixed Fixed $
-      withReaderT (\c -> c & availHeightL .~ h & availWidthL .~ w) $
-        render $ cropToContext p
+setAvailableSize (w, h) p
+    | w <= 0 || h <= 0 = emptyWidget
+    | otherwise =
+        Widget Fixed Fixed $
+          withReaderT (\c -> c & availHeightL .~ h & availWidthL .~ w) $
+            render $ cropToContext p
 
 -- | When drawing the specified widget, set the attribute used for
--- drawing to the one with the specified name. Note that the widget
--- may use further calls to 'withAttr' to change the active drawing
--- attribute, so this only takes effect if nothing in the specified
--- widget invokes 'withAttr'. If you want to prevent that, use
--- 'forceAttr'. Attributes used this way still get merged hierarchically
--- and still fall back to the attribute map's default attribute. If you
--- want to change the default attribute, use 'withDefAttr'.
+-- drawing to the one with the specified name. Note that the widget may
+-- make further changes to the active drawing attribute, so this only
+-- takes effect if nothing in the specified widget invokes 'withAttr'
+-- or otherwise changes the rendering context's attribute setup. If you
+-- want to prevent that, use 'forceAttr'. Attributes used this way still
+-- get merged hierarchically and still fall back to the attribute map's
+-- default attribute. If you want to change the default attribute, use
+-- 'withDefAttr'.
 --
 -- For example:
 --
@@ -883,7 +922,7 @@ setAvailableSize (w, h) p =
 --                         , withAttr "highlight" (str b)
 --                         ]
 --
---    render1 = renderA ("Brick", "fun")
+--    render1 = renderA (\"Brick\", "fun")
 --    render2 = withAttr "warning" render1
 -- @
 --
@@ -927,7 +966,7 @@ modifyDefAttr f p =
 --                         , str " is "
 --                         , withAttr "highlight" (str b) ]
 --
---    render1 = renderA ("Brick", "fun")
+--    render1 = renderA (\"Brick\", "fun")
 --    render2 = withDefAttr "warning" render1
 -- @
 --
@@ -991,6 +1030,17 @@ forceAttr an p =
         c <- getContext
         withReaderT (ctxAttrMapL .~ (forceAttrMap (attrMapLookup an (c^.ctxAttrMapL)))) (render p)
 
+-- | Like 'forceAttr', except that the style of attribute lookups in the
+-- attribute map is preserved and merged with the forced attribute. This
+-- allows for situations where 'forceAttr' would otherwise ignore style
+-- information that is important to preserve.
+forceAttrAllowStyle :: AttrName -> Widget n -> Widget n
+forceAttrAllowStyle an p =
+    Widget (hSize p) (vSize p) $ do
+        c <- getContext
+        let m = c^.ctxAttrMapL
+        withReaderT (ctxAttrMapL .~ (forceAttrMapAllowStyle (attrMapLookup an m) m)) (render p)
+
 -- | Override the lookup of the attribute name 'targetName' to return
 -- the attribute value associated with 'fromName' when rendering the
 -- specified widget.
@@ -1030,6 +1080,31 @@ translateBy off p =
       return $ addResultOffset off
              $ result & imageL %~ (V.translate (off^.locationColumnL) (off^.locationRowL))
 
+-- | Given a widget, translate it to position it relative to the
+-- upper-left coordinates of a reported extent with the specified
+-- positioning offset. If the specified name has no reported extent,
+-- this just draws the specified widget with no special positioning.
+--
+-- This is only useful for positioning something in a higher layer
+-- relative to a reported extent in a lower layer. Any other use is
+-- likely to result in the specified widget being rendered as-is with
+-- no translation. This is because this function relies on information
+-- about lower layer renderings in order to work; using it with a
+-- resource name that wasn't rendered in a lower layer will result in
+-- this being equivalent to @id@.
+--
+-- For example, if you have two layers @topLayer@ and @bottomLayer@,
+-- then a widget drawn in @bottomLayer@ with @reportExtent Foo@ can be
+-- used to relatively position a widget in @topLayer@ with @topLayer =
+-- relativeTo Foo ...@.
+relativeTo :: (Ord n) => n -> Location -> Widget n -> Widget n
+relativeTo n off w =
+    Widget (hSize w) (vSize w) $ do
+        mExt <- lookupReportedExtent n
+        case mExt of
+            Nothing -> render w
+            Just ext -> render $ translateBy (extentUpperLeft ext <> off) w
+
 -- | Crop the specified widget on the left by the specified number of
 -- columns. Defers to the cropped widget for growth policy.
 cropLeftBy :: Int -> Widget n -> Widget n
@@ -1398,10 +1473,10 @@ viewport vpname typ p =
 
       -- If the rendering state includes any scrolling requests for this
       -- viewport, apply those
-      reqs <- lift $ gets $ (^.rsScrollRequestsL)
+      reqs <- lift $ gets (^.rsScrollRequestsL)
       let relevantRequests = snd <$> filter (\(n, _) -> n == vpname) reqs
       when (not $ null relevantRequests) $ do
-          mVp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+          mVp <- lift $ gets (^.viewportMapL.to (M.lookup vpname))
           case mVp of
               Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
               Just vp -> do
@@ -1419,7 +1494,7 @@ viewport vpname typ p =
       -- If the sub-rendering requested visibility, update the scroll
       -- state accordingly
       when (not $ null $ initialResult^.visibilityRequestsL) $ do
-          mVp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+          mVp <- lift $ gets (^.viewportMapL.to (M.lookup vpname))
           case mVp of
               Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
               Just vp -> do
@@ -1432,7 +1507,7 @@ viewport vpname typ p =
 
       -- If the size of the rendering changes enough to make the
       -- viewport offsets invalid, reset them
-      mVp <- lift $ gets $ (^.viewportMapL.to (M.lookup vpname))
+      mVp <- lift $ gets (^.viewportMapL.to (M.lookup vpname))
       vp <- case mVp of
           Nothing -> error $ "BUG: viewport: viewport name " <> show vpname <> " absent from viewport map"
           Just v -> return v
@@ -1513,17 +1588,17 @@ viewport vpname typ p =
 
 -- | The base attribute for scroll bars.
 scrollbarAttr :: AttrName
-scrollbarAttr = "scrollbar"
+scrollbarAttr = attrName "scrollbar"
 
 -- | The attribute for scroll bar troughs. This attribute is a
 -- specialization of @scrollbarAttr@.
 scrollbarTroughAttr :: AttrName
-scrollbarTroughAttr = scrollbarAttr <> "trough"
+scrollbarTroughAttr = scrollbarAttr <> attrName "trough"
 
 -- | The attribute for scroll bar handles. This attribute is a
 -- specialization of @scrollbarAttr@.
 scrollbarHandleAttr :: AttrName
-scrollbarHandleAttr = scrollbarAttr <> "handle"
+scrollbarHandleAttr = scrollbarAttr <> attrName "handle"
 
 maybeClick :: (Ord n)
            => n
