@@ -3,6 +3,8 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TupleSections #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 -- | This module provides a file browser widget that allows users to
 -- navigate directory trees, search for files and directories, and
 -- select entries of interest. For a complete working demonstration of
@@ -141,9 +143,10 @@ module Brick.Widgets.FileBrowser
 where
 
 import qualified Control.Exception as E
-import Control.Monad (forM)
+import Control.Monad (forM, when)
 import Control.Monad.IO.Class (liftIO)
 import Data.Char (toLower, isPrint)
+import Data.Foldable (for_)
 import Data.Maybe (fromMaybe, isJust, fromJust)
 import qualified Data.Foldable as F
 import qualified Data.Text as T
@@ -155,6 +158,7 @@ import Data.List (sortBy, isSuffixOf)
 import qualified Data.Set as Set
 import qualified Data.Vector as V
 import Lens.Micro
+import Lens.Micro.Mtl ((%=))
 import Lens.Micro.TH (lensRules, generateUpdateableOptics)
 import qualified Graphics.Vty as Vty
 import qualified System.Directory as D
@@ -164,7 +168,7 @@ import qualified System.FilePath as FP
 import Text.Printf (printf)
 
 import Brick.Types
-import Brick.AttrMap (AttrName)
+import Brick.AttrMap (AttrName, attrName)
 import Brick.Widgets.Core
 import Brick.Widgets.List
 
@@ -198,6 +202,9 @@ data FileBrowser n =
                 -- change the selection function.
                 }
 
+instance Named (FileBrowser n) n where
+    getName = getName . fileBrowserEntries
+
 -- | File status information.
 data FileStatus =
     FileStatus { fileStatusSize :: Int64
@@ -274,7 +281,7 @@ newFileBrowser :: (FileInfo -> Bool)
                -- executable's current working directory.
                -> IO (FileBrowser n)
 newFileBrowser selPredicate name mCwd = do
-    initialCwd <- case mCwd of
+    initialCwd <- FP.normalise <$> case mCwd of
         Just path -> return path
         Nothing -> D.getCurrentDirectory
 
@@ -312,10 +319,7 @@ selectDirectories :: FileInfo -> Bool
 selectDirectories i =
     case fileInfoFileType i of
         Just Directory -> True
-        Just SymbolicLink ->
-            case fileInfoLinkTargetType i of
-                Just Directory -> True
-                _ -> False
+        Just SymbolicLink -> fileInfoLinkTargetType i == Just Directory
         _ -> False
 
 -- | Set the filtering function used to determine which entries in
@@ -356,10 +360,10 @@ setWorkingDirectory path b = do
             Left (_::E.IOException) -> entries
             Right parent -> parent : entries
 
-    let b' = setEntries allEntries b
-    return $ b' & fileBrowserWorkingDirectoryL .~ path
-                & fileBrowserExceptionL .~ exc
-                & fileBrowserSelectedFilesL .~ mempty
+    return $ setEntries allEntries b
+                 & fileBrowserWorkingDirectoryL .~ path
+                 & fileBrowserExceptionL .~ exc
+                 & fileBrowserSelectedFilesL .~ mempty
 
 parentOf :: FilePath -> IO FileInfo
 parentOf path = getFileInfo ".." $ FP.takeDirectory path
@@ -486,7 +490,7 @@ applyFilterAndSearch :: FileBrowser n -> FileBrowser n
 applyFilterAndSearch b =
     let filterMatch = fromMaybe (const True) (b^.fileBrowserEntryFilterL)
         searchMatch = maybe (const True)
-                            (\search i -> (T.toLower search `T.isInfixOf` (T.pack $ toLower <$> fileInfoSanitizedFilename i)))
+                            (\search i -> T.toLower search `T.isInfixOf` T.pack (toLower <$> fileInfoSanitizedFilename i))
                             (b^.fileBrowserSearchStringL)
         match i = filterMatch i && searchMatch i
         matching = filter match $ b^.fileBrowserLatestResultsL
@@ -575,10 +579,6 @@ fileBrowserCursor b = snd <$> listSelectedElement (b^.fileBrowserEntriesL)
 -- * @Ctrl-f@: 'actionFileBrowserListPageDown'
 -- * @Ctrl-d@: 'actionFileBrowserListHalfPageDown'
 -- * @Ctrl-u@: 'actionFileBrowserListHalfPageUp'
--- * @g@: 'actionFileBrowserListTop'
--- * @G@: 'actionFileBrowserListBottom'
--- * @j@: 'actionFileBrowserListNext'
--- * @k@: 'actionFileBrowserListPrev'
 -- * @Ctrl-n@: 'actionFileBrowserListNext'
 -- * @Ctrl-p@: 'actionFileBrowserListPrev'
 --
@@ -587,132 +587,133 @@ fileBrowserCursor b = snd <$> listSelectedElement (b^.fileBrowserEntriesL)
 -- * @/@: 'actionFileBrowserBeginSearch'
 -- * @Enter@: 'actionFileBrowserSelectEnter'
 -- * @Space@: 'actionFileBrowserSelectCurrent'
+-- * @g@: 'actionFileBrowserListTop'
+-- * @G@: 'actionFileBrowserListBottom'
+-- * @j@: 'actionFileBrowserListNext'
+-- * @k@: 'actionFileBrowserListPrev'
 --
 -- Events handled only in search mode:
 --
 -- * @Esc@, @Ctrl-C@: cancel search mode
 -- * Text input: update search string
 
-actionFileBrowserBeginSearch :: FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserBeginSearch b =
-    return $ updateFileBrowserSearch (const $ Just "") b
+actionFileBrowserBeginSearch :: EventM n (FileBrowser n) ()
+actionFileBrowserBeginSearch =
+    modify $ updateFileBrowserSearch (const $ Just "")
 
-actionFileBrowserSelectEnter :: FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserSelectEnter b =
-    maybeSelectCurrentEntry b
+actionFileBrowserSelectEnter :: EventM n (FileBrowser n) ()
+actionFileBrowserSelectEnter =
+    maybeSelectCurrentEntry
 
-actionFileBrowserSelectCurrent :: FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserSelectCurrent b =
-    selectCurrentEntry b
+actionFileBrowserSelectCurrent :: EventM n (FileBrowser n) ()
+actionFileBrowserSelectCurrent =
+    selectCurrentEntry
 
-actionFileBrowserListPageUp :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListPageUp b = do
-    let old = b ^. fileBrowserEntriesL
-    new <- listMovePageUp old
-    return $ b & fileBrowserEntriesL .~ new
+actionFileBrowserListPageUp :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListPageUp =
+    zoom fileBrowserEntriesL listMovePageUp
 
-actionFileBrowserListPageDown :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListPageDown b = do
-    let old = b ^. fileBrowserEntriesL
-    new <- listMovePageDown old
-    return $ b & fileBrowserEntriesL .~ new
+actionFileBrowserListPageDown :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListPageDown =
+    zoom fileBrowserEntriesL listMovePageDown
 
-actionFileBrowserListHalfPageUp :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListHalfPageUp b = do
-    let old = b ^. fileBrowserEntriesL
-    new <- listMoveByPages (-0.5::Double) old
-    return $ b & fileBrowserEntriesL .~ new
+actionFileBrowserListHalfPageUp :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListHalfPageUp =
+    zoom fileBrowserEntriesL (listMoveByPages (-0.5::Double))
 
-actionFileBrowserListHalfPageDown :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListHalfPageDown b = do
-    let old = b ^. fileBrowserEntriesL
-    new <- listMoveByPages (0.5::Double) old
-    return $ b & fileBrowserEntriesL .~ new
+actionFileBrowserListHalfPageDown :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListHalfPageDown =
+    zoom fileBrowserEntriesL (listMoveByPages (0.5::Double))
 
-actionFileBrowserListTop :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListTop b =
-    return $ b & fileBrowserEntriesL %~ listMoveTo 0
+actionFileBrowserListTop :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListTop =
+    fileBrowserEntriesL %= listMoveTo 0
 
-actionFileBrowserListBottom :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListBottom b = do
+actionFileBrowserListBottom :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListBottom = do
+    b <- get
     let sz = length (listElements $ b^.fileBrowserEntriesL)
-    return $ b & fileBrowserEntriesL %~ listMoveTo (sz - 1)
+    fileBrowserEntriesL %= listMoveTo (sz - 1)
 
-actionFileBrowserListNext :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListNext b =
-    return $ b & fileBrowserEntriesL %~ listMoveBy 1
+actionFileBrowserListNext :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListNext =
+    fileBrowserEntriesL %= listMoveBy 1
 
-actionFileBrowserListPrev :: Ord n => FileBrowser n -> EventM n (FileBrowser n)
-actionFileBrowserListPrev b =
-    return $ b & fileBrowserEntriesL %~ listMoveBy (-1)
+actionFileBrowserListPrev :: Ord n => EventM n (FileBrowser n) ()
+actionFileBrowserListPrev =
+    fileBrowserEntriesL %= listMoveBy (-1)
 
-handleFileBrowserEvent :: (Ord n) => Vty.Event -> FileBrowser n -> EventM n (FileBrowser n)
-handleFileBrowserEvent e b =
+handleFileBrowserEvent :: (Ord n) => Vty.Event -> EventM n (FileBrowser n) ()
+handleFileBrowserEvent e = do
+    b <- get
     if fileBrowserIsSearching b
-    then handleFileBrowserEventSearching e b
-    else handleFileBrowserEventNormal e b
+        then handleFileBrowserEventSearching e
+        else handleFileBrowserEventNormal e
 
 safeInit :: T.Text -> T.Text
 safeInit t | T.length t == 0 = t
            | otherwise = T.init t
 
-handleFileBrowserEventSearching :: (Ord n) => Vty.Event -> FileBrowser n -> EventM n (FileBrowser n)
-handleFileBrowserEventSearching e b =
+handleFileBrowserEventSearching :: (Ord n) => Vty.Event -> EventM n (FileBrowser n) ()
+handleFileBrowserEventSearching e =
     case e of
         Vty.EvKey (Vty.KChar 'c') [Vty.MCtrl] ->
-            return $ updateFileBrowserSearch (const Nothing) b
+            modify $ updateFileBrowserSearch (const Nothing)
         Vty.EvKey Vty.KEsc [] ->
-            return $ updateFileBrowserSearch (const Nothing) b
+            modify $ updateFileBrowserSearch (const Nothing)
         Vty.EvKey Vty.KBS [] ->
-            return $ updateFileBrowserSearch (fmap safeInit) b
-        Vty.EvKey Vty.KEnter [] ->
-            updateFileBrowserSearch (const Nothing) <$>
-                maybeSelectCurrentEntry b
+            modify $ updateFileBrowserSearch (fmap safeInit)
+        Vty.EvKey Vty.KEnter [] -> do
+            maybeSelectCurrentEntry
+            modify $ updateFileBrowserSearch (const Nothing)
         Vty.EvKey (Vty.KChar c) [] ->
-            return $ updateFileBrowserSearch (fmap (flip T.snoc c)) b
+            modify $ updateFileBrowserSearch (fmap (flip T.snoc c))
         _ ->
-            handleFileBrowserEventCommon e b
+            handleFileBrowserEventCommon e
 
-handleFileBrowserEventNormal :: (Ord n) => Vty.Event -> FileBrowser n -> EventM n (FileBrowser n)
-handleFileBrowserEventNormal e b =
+handleFileBrowserEventNormal :: (Ord n) => Vty.Event -> EventM n (FileBrowser n) ()
+handleFileBrowserEventNormal e =
     case e of
         Vty.EvKey (Vty.KChar '/') [] ->
             -- Begin file search
-            actionFileBrowserBeginSearch b
+            actionFileBrowserBeginSearch
         Vty.EvKey Vty.KEnter [] ->
             -- Select file or enter directory
-            actionFileBrowserSelectEnter b
+            actionFileBrowserSelectEnter
         Vty.EvKey (Vty.KChar ' ') [] ->
             -- Select entry
-            actionFileBrowserSelectCurrent b
+            actionFileBrowserSelectCurrent
         _ ->
-            handleFileBrowserEventCommon e b
+            handleFileBrowserEventCommon e
 
-handleFileBrowserEventCommon :: (Ord n) => Vty.Event -> FileBrowser n -> EventM n (FileBrowser n)
-handleFileBrowserEventCommon e b =
+handleFileBrowserEventCommon :: (Ord n) => Vty.Event -> EventM n (FileBrowser n) ()
+handleFileBrowserEventCommon e =
     case e of
         Vty.EvKey (Vty.KChar 'b') [Vty.MCtrl] ->
-            actionFileBrowserListPageUp b
+            actionFileBrowserListPageUp
         Vty.EvKey (Vty.KChar 'f') [Vty.MCtrl] ->
-            actionFileBrowserListPageDown b
+            actionFileBrowserListPageDown
         Vty.EvKey (Vty.KChar 'd') [Vty.MCtrl] ->
-            actionFileBrowserListHalfPageDown b
+            actionFileBrowserListHalfPageDown
         Vty.EvKey (Vty.KChar 'u') [Vty.MCtrl] ->
-            actionFileBrowserListHalfPageUp b
+            actionFileBrowserListHalfPageUp
         Vty.EvKey (Vty.KChar 'g') [] ->
-            actionFileBrowserListTop b
+            actionFileBrowserListTop
         Vty.EvKey (Vty.KChar 'G') [] ->
-            actionFileBrowserListBottom b
+            actionFileBrowserListBottom
         Vty.EvKey (Vty.KChar 'j') [] ->
-            actionFileBrowserListNext b
+            actionFileBrowserListNext
         Vty.EvKey (Vty.KChar 'k') [] ->
-            actionFileBrowserListPrev b
+            actionFileBrowserListPrev
         Vty.EvKey (Vty.KChar 'n') [Vty.MCtrl] ->
-            actionFileBrowserListNext b
+            actionFileBrowserListNext
         Vty.EvKey (Vty.KChar 'p') [Vty.MCtrl] ->
-            actionFileBrowserListPrev b
+            actionFileBrowserListPrev
         _ ->
-            handleEventLensed b fileBrowserEntriesL handleListEvent e
+            zoom fileBrowserEntriesL $ handleListEvent e
+
+markSelected :: FileInfo -> EventM n (FileBrowser n) ()
+markSelected e = fileBrowserSelectedFilesL %= Set.insert (fileInfoFilename e)
 
 -- | If the browser's current entry is selectable according to
 -- @fileBrowserSelectable@, add it to the selection set and return.
@@ -720,30 +721,19 @@ handleFileBrowserEventCommon e b =
 -- directory, set the browser's current path to the selected directory.
 --
 -- Otherwise, return the browser state unchanged.
-maybeSelectCurrentEntry :: FileBrowser n -> EventM n (FileBrowser n)
-maybeSelectCurrentEntry b =
-    case fileBrowserCursor b of
-        Nothing -> return b
-        Just entry ->
-            if fileBrowserSelectable b entry
-            then return $ b & fileBrowserSelectedFilesL %~ Set.insert (fileInfoFilename entry)
-            else case fileInfoFileType entry of
-                Just Directory ->
-                    liftIO $ setWorkingDirectory (fileInfoFilePath entry) b
-                Just SymbolicLink ->
-                    case fileInfoLinkTargetType entry of
-                        Just Directory -> do
-                            liftIO $ setWorkingDirectory (fileInfoFilePath entry) b
-                        _ ->
-                            return b
-                _ ->
-                    return b
+maybeSelectCurrentEntry :: EventM n (FileBrowser n) ()
+maybeSelectCurrentEntry = do
+    b <- get
+    for_ (fileBrowserCursor b) $ \entry ->
+        if fileBrowserSelectable b entry
+        then markSelected entry
+        else when (selectDirectories entry) $
+            put =<< liftIO (setWorkingDirectory (fileInfoFilePath entry) b)
 
-selectCurrentEntry :: FileBrowser n -> EventM n (FileBrowser n)
-selectCurrentEntry b =
-    case fileBrowserCursor b of
-        Nothing -> return b
-        Just e -> return $ b & fileBrowserSelectedFilesL %~ Set.insert (fileInfoFilename e)
+selectCurrentEntry :: EventM n (FileBrowser n) ()
+selectCurrentEntry = do
+    b <- get
+    for_ (fileBrowserCursor b) markSelected
 
 -- | Render a file browser. This renders a list of entries in the
 -- working directory, a cursor to select from among the entries, a
@@ -763,7 +753,7 @@ renderFileBrowser :: (Show n, Ord n)
                   -- ^ The browser to render.
                   -> Widget n
 renderFileBrowser foc b =
-    let maxFilenameLength = maximum $ (length . fileInfoFilename) <$> (b^.fileBrowserEntriesL)
+    let maxFilenameLength = maximum $ length . fileInfoFilename <$> (b^.fileBrowserEntriesL)
         cwdHeader = padRight Max $
                     str $ sanitizeFilename $ fileBrowserWorkingDirectory b
         selInfo = case listSelectedElement (b^.fileBrowserEntriesL) of
@@ -787,7 +777,7 @@ renderFileBrowser foc b =
                                         then ", " <> prettyFileSize (fileStatusSize stat)
                                         else ""
                         in fileTypeLabel (fileStatusFileType stat) <> maybeSize
-            in txt $ (T.pack $ fileInfoSanitizedFilename i) <> ": " <> label
+            in txt $ T.pack (fileInfoSanitizedFilename i) <> ": " <> label
 
         maybeSearchInfo = case b^.fileBrowserSearchStringL of
             Nothing -> emptyWidget
@@ -840,49 +830,49 @@ attrForFileType UnixSocket = fileBrowserUnixSocketAttr
 
 -- | The base attribute for all file browser attributes.
 fileBrowserAttr :: AttrName
-fileBrowserAttr = "fileBrowser"
+fileBrowserAttr = attrName "fileBrowser"
 
 -- | The attribute used for the current directory displayed at the top
 -- of the browser.
 fileBrowserCurrentDirectoryAttr :: AttrName
-fileBrowserCurrentDirectoryAttr = fileBrowserAttr <> "currentDirectory"
+fileBrowserCurrentDirectoryAttr = fileBrowserAttr <> attrName "currentDirectory"
 
 -- | The attribute used for the entry information displayed at the
 -- bottom of the browser.
 fileBrowserSelectionInfoAttr :: AttrName
-fileBrowserSelectionInfoAttr = fileBrowserAttr <> "selectionInfo"
+fileBrowserSelectionInfoAttr = fileBrowserAttr <> attrName "selectionInfo"
 
 -- | The attribute used to render directory entries.
 fileBrowserDirectoryAttr :: AttrName
-fileBrowserDirectoryAttr = fileBrowserAttr <> "directory"
+fileBrowserDirectoryAttr = fileBrowserAttr <> attrName "directory"
 
 -- | The attribute used to render block device entries.
 fileBrowserBlockDeviceAttr :: AttrName
-fileBrowserBlockDeviceAttr = fileBrowserAttr <> "block"
+fileBrowserBlockDeviceAttr = fileBrowserAttr <> attrName "block"
 
 -- | The attribute used to render regular file entries.
 fileBrowserRegularFileAttr :: AttrName
-fileBrowserRegularFileAttr = fileBrowserAttr <> "regular"
+fileBrowserRegularFileAttr = fileBrowserAttr <> attrName "regular"
 
 -- | The attribute used to render character device entries.
 fileBrowserCharacterDeviceAttr :: AttrName
-fileBrowserCharacterDeviceAttr = fileBrowserAttr <> "char"
+fileBrowserCharacterDeviceAttr = fileBrowserAttr <> attrName "char"
 
 -- | The attribute used to render named pipe entries.
 fileBrowserNamedPipeAttr :: AttrName
-fileBrowserNamedPipeAttr = fileBrowserAttr <> "pipe"
+fileBrowserNamedPipeAttr = fileBrowserAttr <> attrName "pipe"
 
 -- | The attribute used to render symbolic link entries.
 fileBrowserSymbolicLinkAttr :: AttrName
-fileBrowserSymbolicLinkAttr = fileBrowserAttr <> "symlink"
+fileBrowserSymbolicLinkAttr = fileBrowserAttr <> attrName "symlink"
 
 -- | The attribute used to render Unix socket entries.
 fileBrowserUnixSocketAttr :: AttrName
-fileBrowserUnixSocketAttr = fileBrowserAttr <> "unixSocket"
+fileBrowserUnixSocketAttr = fileBrowserAttr <> attrName "unixSocket"
 
 -- | The attribute used for selected entries in the file browser.
 fileBrowserSelectedAttr :: AttrName
-fileBrowserSelectedAttr = fileBrowserAttr <> "selected"
+fileBrowserSelectedAttr = fileBrowserAttr <> attrName "selected"
 
 -- | A file type filter for use with 'setFileBrowserEntryFilter'. This
 -- filter permits entries whose file types are in the specified list.
