@@ -4,13 +4,18 @@ module Brick.Widgets.Internal
   , cropToContext
   , cropResultToContext
   , renderDynBorder
+  , renderWidget
   )
 where
 
 import Lens.Micro ((^.), (&), (%~))
-import Control.Monad.Trans.State.Lazy
-import Control.Monad.Trans.Reader
-import Data.Maybe (catMaybes)
+import Lens.Micro.Mtl ((%=))
+import Control.Monad
+import Control.Monad.State.Strict
+import Control.Monad.Reader
+import Data.Maybe (fromMaybe, mapMaybe)
+import qualified Data.Map as M
+import qualified Data.Set as S
 import qualified Graphics.Vty as V
 
 import Brick.Types
@@ -20,7 +25,8 @@ import Brick.Widgets.Border.Style
 import Brick.BorderMap (BorderMap)
 import qualified Brick.BorderMap as BM
 
-renderFinal :: AttrMap
+renderFinal :: (Ord n)
+            => AttrMap
             -> [Widget n]
             -> V.DisplayRegion
             -> ([CursorLocation n] -> Maybe (CursorLocation n))
@@ -31,7 +37,12 @@ renderFinal aMap layerRenders (w, h) chooseCursor rs =
     where
         (layerResults, !newRS) = flip runState rs $ sequence $
             (\p -> runReaderT p ctx) <$>
-            (render <$> cropToContext <$> layerRenders)
+            (\layerWidget -> do
+                result <- render $ cropToContext layerWidget
+                forM_ (result^.extentsL) $ \e ->
+                    reportedExtentsL %= M.insert (extentName e) e
+                return result
+                ) <$> reverse layerRenders
 
         ctx = Context { ctxAttrName = mempty
                       , availWidth = w
@@ -50,14 +61,16 @@ renderFinal aMap layerRenders (w, h) chooseCursor rs =
                       , ctxHScrollBarClickableConstr = Nothing
                       , ctxVScrollBarClickableConstr = Nothing
                       }
-        pic = V.picForLayers $ uncurry V.resize (w, h) <$> (^.imageL) <$> layerResults
+
+        layersTopmostFirst = reverse layerResults
+        pic = V.picForLayers $ V.resize w h <$> (^.imageL) <$> layersTopmostFirst
 
         -- picWithBg is a workaround for runaway attributes.
         -- See https://github.com/coreyoconnor/vty/issues/95
         picWithBg = pic { V.picBackground = V.Background ' ' V.defAttr }
 
-        layerCursors = (^.cursorsL) <$> layerResults
-        layerExtents = reverse $ (^.extentsL) <$> layerResults
+        layerCursors = (^.cursorsL) <$> layersTopmostFirst
+        layerExtents = reverse $ (^.extentsL) <$> layersTopmostFirst
         theCursor = chooseCursor $ concat layerCursors
 
 -- | After rendering the specified widget, crop its result image to the
@@ -78,7 +91,7 @@ cropImage :: Context n -> V.Image -> V.Image
 cropImage c = V.crop (max 0 $ c^.availWidthL) (max 0 $ c^.availHeightL)
 
 cropCursors :: Context n -> [CursorLocation n] -> [CursorLocation n]
-cropCursors ctx cs = catMaybes $ cropCursor <$> cs
+cropCursors ctx cs = mapMaybe cropCursor cs
     where
         -- A cursor location is removed if it is not within the region
         -- described by the context.
@@ -92,7 +105,7 @@ cropCursors ctx cs = catMaybes $ cropCursor <$> cs
                ]
 
 cropExtents :: Context n -> [Extent n] -> [Extent n]
-cropExtents ctx es = catMaybes $ cropExtent <$> es
+cropExtents ctx es = mapMaybe cropExtent es
     where
         -- An extent is cropped in places where it is not within the
         -- region described by the context.
@@ -143,3 +156,36 @@ renderDynBorder db = V.char (dbAttr db) $ getBorderChar $ dbStyle db
             Edges True  True  False True  -> bsIntersectL
             Edges True  True  True  False -> bsIntersectR
             Edges True  True  True  True  -> bsIntersectFull
+
+-- | This function provides a simplified interface to rendering a list
+-- of 'Widget's as a 'V.Picture' outside of the context of an 'App'.
+-- This can be useful in a testing setting but isn't intended to be used
+-- for normal application rendering. The API is deliberately narrower
+-- than the main interactive API and is not yet stable. Use at your own
+-- risk.
+--
+-- Consult the [Vty library documentation](https://hackage.haskell.org/package/vty)
+-- for details on how to output the resulting 'V.Picture'.
+renderWidget :: (Ord n)
+             => Maybe AttrMap
+             -- ^ Optional attribute map used to render. If omitted,
+             -- an empty attribute map with the terminal's default
+             -- attribute will be used.
+             -> [Widget n]
+             -- ^ The widget layers to render, topmost first.
+             -> V.DisplayRegion
+             -- ^ The size of the display region in which to render the
+             -- layers.
+             -> V.Picture
+renderWidget mAttrMap layerRenders region = pic
+    where
+        initialRS = RS { viewportMap = M.empty
+                       , rsScrollRequests = []
+                       , observedNames = S.empty
+                       , renderCache = mempty
+                       , clickableNames = []
+                       , requestedVisibleNames_ = S.empty
+                       , reportedExtents = mempty
+                       }
+        am = fromMaybe (attrMap V.defAttr []) mAttrMap
+        (_, pic, _, _) = renderFinal am layerRenders region (const Nothing) initialRS
